//***************************************************************************
// Group Linslot / Linux - Slotrace Manager
// File arduino.cc
// Date 16.11.07 - Jrg Wendel
// This code is distributed under the terms and conditions of the
// GNU GENERAL PUBLIC LICENSE. See the file COPYING for details.
//***************************************************************************

//***************************************************************************
// Include
//***************************************************************************

#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>

#include <arduino.hpp>

//***************************************************************************
// Object
//***************************************************************************

Arduino::Arduino()
   : IoInterface()
{
   fdDevice = 0;
   opened = no;
   readTimeout = 60;
   writeTimeout = 60;
   outValue = 0;
   messageSize = 0;
   *message = 0;

   bzero(&oldtio, sizeof(oldtio));
}

Arduino::~Arduino()
{
   close();
}

//***************************************************************************
// Settings
//***************************************************************************

int Arduino::setTimeout(unsigned int timeout)
{
   readTimeout = timeout / 1000;
   return success;
}

int Arduino::setWriteTimeout(unsigned int timeout)
{
   writeTimeout = timeout / 1000;
   return success;
}

//***************************************************************************
// Open/Close Device
//***************************************************************************

int Arduino::open(const char* dev)
{
   struct termios newtio;

   if (!dev)
      dev = deviceName;
   else
      strcpy(deviceName, dev);

   if (!deviceName || !*deviceName)
      return fail;

   tell(eloAlways, "Try to open io device '%s'", deviceName);

   deviceMutex.lock();

   if ((fdDevice = ::open(deviceName, O_RDWR | O_NOCTTY)) < 0)
   {
      deviceMutex.unlock();
      fdDevice = 0;

      tell(eloAlways, "Error: Opening '%s' failed", deviceName);

      return fail;
   }

   // configure serial line

   tcgetattr(fdDevice, &oldtio);
   bzero(&newtio, sizeof(newtio));

   // B9600    x
   // B19200   x
   // B38400   x
   // B57600   x
   // B115200

   newtio.c_cflag = B57600 | CS8 | CLOCAL | CREAD;
   newtio.c_iflag = IGNPAR | ICRNL; // PARMRK dont work for bin data !;
   newtio.c_oflag = 0;
   newtio.c_lflag = 0;
   tcflush(fdDevice, TCIFLUSH);
   tcsetattr(fdDevice, TCSANOW, &newtio);

   deviceMutex.unlock();

   flush();

   // first ask about the boardtime

   initBoardTime();
   recordGhostCar(na, na);       // switch off ghostcar recording 
   sendCommand(cGetInputs);      // get actual inpiut state

   tell(eloAlways, "Initialized io device successfully");

   opened = yes;

   return success;
}

int Arduino::close()
{
   // restore the old settings and close

   if (fdDevice)
   {
      tell(eloDebug, "Debug: Closing connection");

      QMutexLocker locker(&deviceMutex);

      tcsetattr(fdDevice, TCSANOW, &oldtio);
      ::close(fdDevice);
      fdDevice = 0;
   }

   opened = no;

   return success;
}

//***************************************************************************
// Re Open
//***************************************************************************

int Arduino::reopen(const char* dev)
{
   close();
   
   return open(dev);
}

//***************************************************************************
// Flush
//***************************************************************************

int Arduino::flush()
{
   byte command;
   
   if (fdDevice)
   {
      while (look(command) == success)
         ;
   }

   return success;
}

//***************************************************************************
// Init Board Time
//***************************************************************************

void Arduino::initBoardTime()
{
   byte command;
   int tries = 0, cnt = 0;

   // ask about the boardtime

   tell(eloDebug, "Debug: Try to get the board time");

   while (command != cBoardTime)
   {
      if (!(tries % 100))
      {
         cnt++;
         sendCommand(cGetTime);
      }

      usleep(1000);
      look(command);
      tries++;
   }

   tell(eloAlways, "Board time initialized after (%d) tries!", cnt);
}

//***************************************************************************
// Write IO Bit
//***************************************************************************

int Arduino::writeBit(int bit, int state)
{
   if (!fdDevice || bit == na)
      return fail;

   setBitTo(outValue, bit, state);

   tell(eloDebug, "Debug: Write bit (%d) value (%d)", 
        bit, state);

   DigitalOutputBit digital;
   digital.bit = bit;
   digital.state = state;

   sendCommand(cDigitalOutBit, &digital, sizeof(DigitalOutputBit));

   return success;
}

//***************************************************************************
// Read Out Bit
//
//  - Return the actual Value of the given 'out' Bit,
//    using the cached value of last write!
//***************************************************************************

int Arduino::readOutBit(int bit)
{
   char buf[100+TB];
   
   tell(eloDebug2, "Debug: out value is '%s'", toBinStr(outValue, buf));

   return isBit(outValue, bit);
}

//***************************************************************************
// Analog Write
//***************************************************************************

void Arduino::analogWrite(int bit, byte value)
{
   if (fdDevice)
   {
      AnalogOutput analogOut;

      analogOut.bit = bit;
      analogOut.value = value;

      sendCommand(cAnalogOut, &analogOut, sizeof(AnalogOutput));
   }
}

//***************************************************************************
// Recorst Ghost Car
//***************************************************************************

void Arduino::recordGhostCar(int vBit, int iBit)
{
   if (fdDevice)
   {
      RecordGhostCar rgc;

      rgc.cycle = getGcScale();
      rgc.bits = vBit | (iBit << 4);

      tell(eloAlways, "send 0x%x", rgc.bits);
      sendCommand(cRecordGhostCar, &rgc, sizeof(RecordGhostCar));

      // analogWrite(outputBits[bitPmwOutSlot1].bit, 254);
   }
}

//***************************************************************************
// Read Value
//***************************************************************************

int Arduino::look(byte& command)
{
   int cmd;

   *message = 0;
   messageSize = 0;
   command = cNone;

   if (!fdDevice)
      return fail;

   if (_look() != success)
   {
      // no input data pending

      return fail;
   }

   tell(eloDebug2, "Debug: look() succeeded, reading command");

   if ((cmd = receiveCommand()) == fail)
   {
      tell(eloAlways, "Error: receiveCommand() failed");
      return fail;
   }

   command = cmd;

   if (command == cBoardTime)
   {
      unsigned int boardTime;
      DigitalInput* boardTm;
      boardTm = (DigitalInput*)message;
      
      tell(eloAlways, "<- BoardTime: (%ums)", boardTm->time);
      
      boardTime = boardTm->time;
      tvNow(&tvBoardStartTime);
      tvBoardStartTime = subMs2Tv(tvBoardStartTime, boardTime);

      *message = 0;
      messageSize = 0;

      return fail;
   }

   return success;
}

//***************************************************************************
// Send Command
//***************************************************************************

int Arduino::sendCommand(byte command, void* line, byte size)
{
   QMutexLocker locker(&deviceMutex);

   // tell(eloDebug2, "-> [%s]", cmd);

   byte cmd = protocol | command & 0x0F;

   ::write(fdDevice, &cmd, 1);
   ::write(fdDevice, &size, 1);

   if (line)
      ::write(fdDevice, line, size);

   return done;
}

//***************************************************************************
// Look
//***************************************************************************

int Arduino::_look()
{
   byte b;

   QMutexLocker locker(&deviceMutex);

   if (::read(fdDevice, &b, 1) <= 0)
      return fail;

   if ((b & 0xF0) != protocol)
   {
      tell(eloAlways, "Info: Protocol violation, skipping byte [0x%x]!", b);
      return fail;
   }

   command = b & 0x0F;

   return success;
}

//***************************************************************************
// Read Command
//***************************************************************************

int Arduino::receiveCommand()
{
   int count = 0;
   int res;

   QMutexLocker locker(&deviceMutex);

   // for protocol

   while ((res = ::read(fdDevice, &messageSize, 1)) <= 0)
   {
      if (res < 0)
      {
         tell(eloAlways, "Error: read failed!");
         return fail;
      }
      
      usleep(100);
   }

   tell(eloDebug, "Debug: Got command (0x%X) expected size is (%d)", command, messageSize);

   while (count < messageSize)
   {
      res = ::read(fdDevice, message+count, messageSize-count);

      tell(eloDebug2, "Debug: read() result was (%d)", res);

      if (res <= 0)
         usleep(100);
      else
         count += res;
   }
   
   tell(eloDebug2, "Debug: read() (%d) bytes", count);

   return command;
}
