//***************************************************************************
// Group Common Stuff
// File common.cc
// Date 04.11.06 - Jrg Wendel
// This code is distributed under the terms and conditions of the
// GNU GENERAL PUBLIC LICENSE. See the file COPYING for details.
//***************************************************************************

//***************************************************************************
// Includes
//***************************************************************************

#include <sys/time.h>
#include <stdarg.h>
#include <time.h>
#include <stdio.h>
#include <string.h>

#include <common.hpp>

//***************************************************************************
// Globals
//***************************************************************************

int theEloquence;

//***************************************************************************
// Tell
//***************************************************************************

int tell(int eloquence, const char* format, ...)
{
   const int sizeTime = 8;        // "12:12:34"
   const int sizeMSec = 4;        // ",142"
   const int sizeHeader = sizeTime + sizeMSec + 1;
   const int maxBuf = 1000;

   struct timeval tp;
   char buf[maxBuf+TB];
   va_list ap;
   time_t now;

   va_start(ap, format);

   if (theEloquence >= eloquence)
   {
      time(&now);
      gettimeofday(&tp, 0);

      vsnprintf(buf + sizeHeader, maxBuf - sizeHeader, format, ap);
      strftime(buf, sizeTime+1, "%H:%M:%S", localtime(&now));

      sprintf(buf+sizeTime, ",%3.3ld", tp.tv_usec / 1000);
      
      buf[sizeHeader-1] = ' ';
      printf("%s\n", buf);
   }

   va_end(ap);
   
   return 0;
}

//***************************************************************************
// Not Null
//***************************************************************************

const char* notNull(const char* s)
{ 
   return s ? s : ""; 
}

//***************************************************************************
// Class SlotService
//***************************************************************************
//***************************************************************************
// Is Bit
//***************************************************************************

int SlotService::isBit(unsigned int value, int bit)
{
   return value & (1 << bit);
}

//***************************************************************************
// Toggle Bit
//***************************************************************************

void SlotService::toggleBit(unsigned int& value, int bit)
{
   setBitTo(value, bit, !isBit(value, bit));
}

//***************************************************************************
// Set Bit To
//***************************************************************************

void SlotService::setBitTo(unsigned int& value, int bit, int state)
{
   if (state)
      value |= 1 << bit;
   else
      value &= ~(1 << bit);
}

//***************************************************************************
// To Bin Str
//***************************************************************************

const char* SlotService::toBinStr(unsigned int value, char* mask, int bits)
{
   int bytes = bits / 8;
   int bit = 1;

   if (bytes * 8 < bits)
      bytes++;

   for (int i = 0, b = 0; b < bits; i++, b++)
   {
      if (!(b % 8) && b)
      {
         mask[bits+bytes-i-2] = ':';
         i++;
      }
      
      mask[bits+bytes-i-2] = (value & bit) ? '1' : '0';     
      bit = bit << 1;
   }

   mask[bits+bytes-1] = 0;

   return mask;
}

//***************************************************************************
// kmh
//***************************************************************************

double SlotService::kmh(double meter, double usec)
{
   return ((double)meter/1000L)  /  
      ((double)usec / 1000000L / 60L / 60L);
}

//***************************************************************************
// Init TV
//***************************************************************************

void SlotService::tvNull(timeval* tv)
{
   memset(tv, 0, sizeof(timeval));
}

timeval* SlotService::tvNow(timeval* tv)
{
   gettimeofday(tv, 0);
   return tv;
}

//***************************************************************************
// Add/Sub from timeval
//***************************************************************************

timeval SlotService::addMs2Tv(timeval tv, unsigned long milli)
{
   timeval tvSum;
   unsigned long long usec = tv.tv_usec + (milli % 1000) * 1000;

   tvSum.tv_usec = usec % 1000000;
   tvSum.tv_sec  = tv.tv_sec + (milli / 1000) + (usec / 1000000);

   return tvSum;
}

timeval SlotService::subMs2Tv(timeval tv, unsigned long milli)
{
   unsigned long long sec = (milli / 1000);
   long long usec = (milli % 1000) * 1000;

   tv.tv_sec -= sec;
   usec = tv.tv_usec - usec;

   if (usec > 0)
      tv.tv_usec = usec;
   else
   {
      tv.tv_sec--;
      tv.tv_usec = 1000000 + usec;
   }

   return tv;
}

//***************************************************************************
// Elapsed in Seconds
//***************************************************************************

long long SlotService::elapsed(const timeval* tpFrom, const timeval* tpNow)
{
   long long sec = tpNow->tv_sec - tpFrom->tv_sec;
   long long usec = sec*1000000 + tpNow->tv_usec;

   usec -= tpFrom->tv_usec;

   return usec;
}

//***************************************************************************
// Seconds String
//***************************************************************************

const char* SlotService::secStr(long usec, char* buf)
{
   sprintf(buf, "%2ld.%06ld", usec/1000000, usec%1000000);

   return buf;
}

//***************************************************************************
// To Str
//***************************************************************************

const char* SlotService::toStr(int value)
{
   static char buf[10+TB];

   sprintf(buf, "%d", value);

   return buf;
}

//***************************************************************************
// Definition of Input/Output Bits
//
//   - set unused features to 'na'
//***************************************************************************

SlotService::InputDefinition SlotService::inputBits[] =
{ 
   // bit, mode,              function

   {   4, teFalling  },   //  bitStartKey

   {  16, teRising   },   //  bitPanicKey1 
   {  17, teRising   },   //  bitPanicKey2  
   {  18, teRising   },   //  bitPanicKey3 
   {  19, teRising   },   //  bitPanicKey4 

   {  20, teFalling  },   //  bitIrSlot1
   {  21, teFalling  },   //  bitIrSlot2
   {  22, teFalling  },   //  bitIrSlot3   
   {  23, teFalling  },   //  bitIrSlot4   

   {  24, teFalling  },   //  bitFuelStartSlot1
   {  25, teFalling  },   //  bitFuelStartSlot2
   {  26, teFalling  },   //  bitFuelStartSlot3
   {  27, teFalling  },   //  bitFuelStartSlot4

   {  28, teFalling  },   //  bitFuelEndSlot1
   {  29, teFalling  },   //  bitFuelEndSlot2
   {  30, teFalling  },   //  bitFuelEndSlot3
   {  31, teFalling  }    //  bitFuelEndSlot4
};

SlotService::OutputDefinition SlotService::outputBits[] =
{
   // bit, mode,              function

   {  16, omNormal   },   //  bitLightRed1
   {  17, omNormal   },   //  bitLightRed2     
   {  18, omNormal   },   //  bitLightRed3     
   {  19, omNormal   },   //  bitLightRed4     
   {  20, omNormal   },   //  bitLightRed5     
   {  21, omNormal   },   //  bitLightGreen    

   {  24, omNormal   },   //  bitLightYellow1  
   {  25, omNormal   },   //  bitLightYellow2  
   {  26, omNormal   },   //  bitLightYellow3  
   {  27, omNormal   },   //  bitLightYellow4  

   {  28, omNormal   },   //  bitPowerSlot1
   {  29, omNormal   },   //  bitPowerSlot2
   {  30, omNormal   },   //  bitPowerSlot3
   {  31, omNormal   },   //  bitPowerSlot4

   {   5, omPwm      },   //  bitPmwOutSlot1
   {   6, omPwm      },   //  bitPmwOutSlot2
};

const char* SlotService::inputFunctions[] =
{
   "StartKey",
            
   "PanicKey1",
   "PanicKey2",
   "PanicKey3",
   "PanicKey4",
   
   "IrSlot1",
   "IrSlot2",
   "IrSlot3",
   "IrSlot4",
   
   "FuelStartSlot1",
   "FuelStartSlot2",
   "FuelStartSlot3",
   "FuelStartSlot4",
   
   "FuelEndSlot1",
   "FuelEndSlot2",
   "FuelEndSlot3",
   "FuelEndSlot4",
};

const char* SlotService::outputFunctions[] =
{
   "LightRed1",
   "LightRed2",
   "LightRed3",  
   "LightRed4",  
   "LightRed5", 
   "LightGreen", 
   
   "LightYellow1",
   "LightYellow2",
   "LightYellow3",
   "LightYellow4",
   
   "PowerSlot1", 
   "PowerSlot2",
   "PowerSlot3",
   "PowerSlot4",
   
   "PmwOutSlot1",
   "PmwOutSlot2"  
};

const char* SlotService::inputFunctionName(int fct)
{
   if (fct < 0 || fct > bitInputCount)
      return "<unknown>";
   
   return inputFunctions[fct];
}

const char* SlotService::outputFunctionName(int fct)
{
   if (fct < 0 || fct > bitOutputCount)
      return "<unknown>";
   
   return outputFunctions[fct];
}


